#include "opendrive_link.h"
#include "opendrive_road_node.h"
#include "exceptions.h"

COpendriveLink::COpendriveLink()
{
  this->prev=NULL;
  this->next=NULL;
  this->segment=NULL;
  this->lane=NULL;
  this->spline=NULL;
  this->mark=OD_MARK_NONE;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

COpendriveLink::COpendriveLink(const COpendriveLink &object)
{
  this->prev=object.prev;
  this->next=object.next;
  this->segment=object.segment;
  this->lane=object.lane;
  this->spline=new CG2Spline(*object.spline);
  this->mark=object.mark;
  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
}

void COpendriveLink::set_prev(COpendriveRoadNode *node)
{
  this->prev=node;
}

void COpendriveLink::set_next(COpendriveRoadNode *node)
{
  this->next=node;
}

void COpendriveLink::set_road_mark(opendrive_mark_t mark)
{
  this->mark=mark;
}

void COpendriveLink::set_parent_segment(COpendriveRoadSegment *segment)
{
  this->segment=segment;
}

void COpendriveLink::set_parent_lane(COpendriveLane *lane)
{
  this->lane=lane;
}

void COpendriveLink::set_resolution(double res)
{
  this->resolution=res;

  if(this->spline!=NULL)
    this->spline->generate(res);
}

void COpendriveLink::set_scale_factor(double scale)
{
  TPoint spline_start,spline_end;

  if(this->spline!=NULL)
  {
    this->spline->get_start_point(spline_start);
    spline_start.x*=this->scale_factor/scale;
    spline_start.y*=this->scale_factor/scale;
    spline_start.curvature*=scale/this->scale_factor;
    this->spline->set_start_point(spline_start);
    this->spline->get_end_point(spline_end);
    spline_end.x*=this->scale_factor/scale;
    spline_end.y*=this->scale_factor/scale;
    spline_end.curvature*=scale/this->scale_factor;
    this->spline->set_end_point(spline_end);
    this->spline->generate(this->resolution);
  }

  this->scale_factor=scale;
}

void COpendriveLink::generate(double start_curvature,double end_curvature)
{
  TOpendriveWorldPose node_start,node_end;
  TPoint start,end;

  node_start=this->prev->get_pose();
  start.x=node_start.x;
  start.y=node_start.y;
  start.heading=node_start.heading;
  start.curvature=start_curvature;
  node_end=this->next->get_pose();
  end.x=node_end.x;
  end.y=node_end.y;
  end.heading=node_end.heading;
  end.curvature=end_curvature;
  this->spline=new CG2Spline(start,end);
  this->spline->generate(this->resolution);
}

void COpendriveLink::update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs)
{
  if(node_refs.find(this->prev)!=node_refs.end())
    this->prev=node_refs[this->prev];
  if(node_refs.find(this->next)!=node_refs.end())
    this->next=node_refs[this->next];
  if(segment_refs.find(this->segment)!=segment_refs.end())
    this->segment=segment_refs[this->segment];
  if(lane_refs.find(this->lane)!=lane_refs.end())
    this->lane=lane_refs[this->lane];
}

bool COpendriveLink::clean_references(node_up_ref_t &refs)
{
  if(refs.find(this->prev)!=refs.end())
    this->prev=refs[this->prev];
  else if(!this->prev->updated(refs))
    return true;
  if(refs.find(this->next)!=refs.end())
    this->next=refs[this->next];
  else if(!this->next->updated(refs))
    return true;

  return false;
}

void COpendriveLink::update_start_pose(TOpendriveWorldPose &start,double curvature)
{
  TPoint start_pose;

  if(this->spline!=NULL)
  {
    start_pose.x=start.x;
    start_pose.y=start.y;
    start_pose.heading=start.heading;
    start_pose.curvature=curvature;
    this->spline->set_start_point(start_pose);
    this->spline->generate(this->resolution);
  }
}

void COpendriveLink::update_end_pose(TOpendriveWorldPose &end,double curvature)
{
  TPoint end_pose;

  if(this->spline!=NULL)
  {
    end_pose.x=end.x;
    end_pose.y=end.y;
    end_pose.heading=end.heading;
    end_pose.curvature=curvature;
    this->spline->set_end_point(end_pose);
    this->spline->generate(this->resolution);
  }
}

const COpendriveRoadNode &COpendriveLink::get_prev(void) const
{
  return *this->prev;
}

const COpendriveRoadNode &COpendriveLink::get_next(void) const
{
  return *this->next;
}

opendrive_mark_t COpendriveLink::get_road_mark(void) const
{
  return this->mark;
}

const COpendriveRoadSegment &COpendriveLink::get_parent_segment(void) const
{
  return *this->segment;
}

const COpendriveLane &COpendriveLink::get_parent_lane(void) const
{
  std::stringstream error;

  if(this->lane!=NULL)
    return *this->lane;
  else
  {
    error << "Link from node " << this->prev->get_index() << " to node " << this->next->get_index() << " has no parent lane";
    throw CException(_HERE_,error.str());
  }
}

double COpendriveLink::get_resolution(void) const
{
  return this->resolution;
}

double COpendriveLink::get_scale_factor(void) const
{
  return this->scale_factor;
}

double COpendriveLink::find_closest_world_pose(TOpendriveWorldPose &world,TOpendriveWorldPose &pose)
{
  TPoint spline_pose;
  double length;

  if(this->spline!=NULL)
  {
    length=this->spline->find_closest_point(world.x,world.y,spline_pose);
    pose.x=spline_pose.x;
    pose.y=spline_pose.y;
    pose.heading=normalize_angle(spline_pose.heading);
  }
  else
    length=std::numeric_limits<double>::max();

  return length;
}

void COpendriveLink::get_trajectory(std::vector<double> &x, std::vector<double> &y,double start_length, double end_length) const
{
  std::vector<double> curvature,heading;
  CG2Spline *partial_spline=NULL;
  
  if(start_length!=0.0 || end_length!=-1.0)// get partial spline
  {
    partial_spline=new CG2Spline;
    this->spline->get_part(partial_spline,start_length,end_length);
    partial_spline->evaluate_all(x,y,curvature,heading);
    delete partial_spline;
  }
  else
    this->spline->evaluate_all(x,y,curvature,heading);
}

void COpendriveLink::get_trajectory(std::vector<double> &x, std::vector<double> &y, std::vector<double> &yaw,double start_length, double end_length) const
{
  std::vector<double> curvature;
  CG2Spline *partial_spline=NULL;

  if(start_length!=0.0 || end_length!=-1.0)// get partial spline
  {
    partial_spline=new CG2Spline;
    this->spline->get_part(partial_spline,start_length,end_length);
    partial_spline->evaluate_all(x,y,curvature,yaw);
    delete partial_spline;
  }
  else
    this->spline->evaluate_all(x,y,curvature,yaw);
}

double COpendriveLink::get_length(void) const
{
  if(this->spline!=NULL)
    return this->spline->get_length();
  else
    return 0.0;
}

TOpendriveWorldPose COpendriveLink::get_pose_at(double length)
{
  TPoint spline_pose;
  TOpendriveWorldPose world_pose={0};

  if(this->spline!=NULL)
  {
    spline_pose=this->spline->evaluate(length);
    world_pose.x=spline_pose.x;
    world_pose.y=spline_pose.y;
    world_pose.heading=spline_pose.heading;
  }

  return world_pose;
}

double COpendriveLink::get_curvature_at(double length)
{
  TPoint spline_pose;
  double curvature=0.0;

  if(this->spline!=NULL)
  {
    spline_pose=this->spline->evaluate(length);
    curvature=spline_pose.curvature;
  }

  return curvature;
}

std::ostream& operator<<(std::ostream& out, COpendriveLink &link)
{
  out << "        Previous node: " << link.get_prev().get_index() << std::endl;
  out << "        Next node: " << link.get_next().get_index() << std::endl;
  if(link.lane!=NULL)
    out << "        Parent segment: " << link.get_parent_segment().get_name() << " (lane " << link.get_parent_lane().get_id() << ")" << std::endl;
  else
    out << "        Parent segment: " << link.get_parent_segment().get_name() << " (no lane)" << std::endl;
  out << "        Road mark: ";
  switch(link.get_road_mark())
  {
    case OD_MARK_NONE:
      out << "no mark" << std::endl;
      break;
    case OD_MARK_SOLID:
      out << "solid" << std::endl;
      break;
    case OD_MARK_BROKEN:
      out << "broken" << std::endl;
      break;
    case OD_MARK_SOLID_SOLID:
      out << "solid solid" << std::endl;
      break;
    case OD_MARK_SOLID_BROKEN:
      out << "solid broken" << std::endl;
      break;
    case OD_MARK_BROKEN_SOLID:
      out << "broken solid" << std::endl;
      break;
    case OD_MARK_BROKEN_BROKEN:
      out << "broken broken" << std::endl;
      break;
  }

  return out;
}

COpendriveLink::~COpendriveLink()
{
  this->prev=NULL;
  this->next=NULL;
  this->segment=NULL;
  this->lane=NULL;
  if(this->spline!=NULL)
  {
    delete this->spline;
    this->spline=NULL;
  }
  this->mark=OD_MARK_NONE;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

