#include "opendrive_road_segment.h"
#include "exceptions.h"
#include <math.h>

COpendriveRoadSegment::COpendriveRoadSegment()
{
  this->name="";
  this->id=-1;
  this->center_mark=OD_MARK_NONE;
  this->lanes.clear();
  this->num_left_lanes=0;
  this->num_right_lanes=0;
  this->nodes.clear();
  this->signals.clear();
  this->objects.clear();
  this->connecting.clear();
  this->min_road_length=DEFAULT_MIN_ROAD_LENGTH;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->geometries.clear();
}

COpendriveRoadSegment::COpendriveRoadSegment(const COpendriveRoadSegment &object)
{
  this->name=object.name;
  this->id=object.id;
  this->lanes=object.lanes;
  this->num_right_lanes=object.num_right_lanes;
  this->num_left_lanes=object.num_left_lanes;
  this->nodes=object.nodes;
  this->parent_road=object.parent_road;
  this->signals.clear();
  this->signals.resize(object.signals.size());
  for(unsigned int i=0;i<object.signals.size();i++)
    this->signals[i]=new COpendriveSignal(*object.signals[i]);
  this->objects.resize(object.objects.size());
  for(unsigned int i=0;i<object.signals.size();i++)
    this->objects[i]=new COpendriveObject(*object.objects[i]);
  this->connecting=object.connecting;
  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
  this->min_road_length=object.min_road_length;
  this->center_mark=object.center_mark;
  this->geometries.resize(object.geometries.size());
  for(unsigned int i=0;i<object.geometries.size();i++)
  {
    this->geometries[i].opendrive=object.geometries[i].opendrive->clone();
    this->geometries[i].spline=new CG2Spline(*object.geometries[i].spline);
  }
}

void COpendriveRoadSegment::free(void)
{
  this->lanes.clear();
  this->num_left_lanes=0;
  this->num_right_lanes=0;
  for(unsigned int i=0;i<this->signals.size();i++)
  {
    delete this->signals[i];
    this->signals[i]=NULL;
  }
  this->nodes.clear();
  this->signals.clear();
  for(unsigned int i=0;i<this->objects.size();i++)
  {
    delete this->objects[i];
    this->objects[i]=NULL;
  }
  this->objects.clear();
  this->connecting.clear();
  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    delete this->geometries[i].opendrive;
    this->geometries[i].opendrive=NULL;
    delete this->geometries[i].spline;
    this->geometries[i].spline=NULL;
  }
  this->geometries.clear();
}

void COpendriveRoadSegment::set_resolution(double res)
{
  std::map<int,COpendriveLane *>::const_iterator lane_it;

  this->resolution=res;

  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();++lane_it)
    lane_it->second->set_resolution(res);

  for(unsigned int i=0;i<this->geometries.size();i++)
    this->geometries[i].spline->generate(res);
}

void COpendriveRoadSegment::set_scale_factor(double scale)
{
  std::map<int,COpendriveLane *>::const_iterator lane_it;
  TPoint spline_start,spline_end;

  this->scale_factor=scale;

  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();++lane_it)
    lane_it->second->set_scale_factor(scale);
  for(unsigned int i=0;i<this->signals.size();i++)
    this->signals[i]->set_scale_factor(scale);
  for(unsigned int i=0;i<this->objects.size();i++)
    this->objects[i]->set_scale_factor(scale);

  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    this->geometries[i].opendrive->set_scale_factor(scale);
    this->geometries[i].spline->get_start_point(spline_start);
    spline_start.x*=this->scale_factor/scale;
    spline_start.y*=this->scale_factor/scale;
    spline_start.curvature*=scale/this->scale_factor;
    this->geometries[i].spline->set_start_point(spline_start);
    this->geometries[i].spline->get_end_point(spline_end);
    spline_end.x*=this->scale_factor/scale;
    spline_end.y*=this->scale_factor/scale;
    spline_end.curvature*=scale/this->scale_factor;
    this->geometries[i].spline->set_end_point(spline_end);
    this->geometries[i].spline->generate(this->resolution);
  }
}

void COpendriveRoadSegment::set_min_road_length(double length)
{
  this->min_road_length=length;
}

void COpendriveRoadSegment::set_parent_road(COpendriveRoad *parent)
{
  this->parent_road=parent;
}

void COpendriveRoadSegment::set_name(std::string &name)
{
  this->name=name;
}

void COpendriveRoadSegment::set_id(unsigned int id)
{
  this->id=id;
}

void COpendriveRoadSegment::set_center_mark(opendrive_mark_t mark)
{
  this->center_mark=mark;
}

void COpendriveRoadSegment::add_geometries(OpenDRIVE::road_type &road_info)
{
  planView::geometry_iterator geom_info;
  TOpendriveRoadSegmentGeometry geometry;
  TOpendriveWorldPose start_pose,end_pose;
  TPoint spline_start,spline_end;
  double start_curvature,end_curvature;
  std::stringstream error;

  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    delete this->geometries[i].opendrive;
    delete this->geometries[i].spline;
  }
  this->geometries.clear();
  for(geom_info=road_info.planView().geometry().begin(); geom_info!=road_info.planView().geometry().end(); ++geom_info)
  {
    // import geometry
    if(geom_info->line().present())
      geometry.opendrive=new COpendriveLine();
    else if(geom_info->spiral().present())
      geometry.opendrive=new COpendriveSpiral();
    else if(geom_info->arc().present())
      geometry.opendrive=new COpendriveArc();
    else if(geom_info->paramPoly3().present())
      geometry.opendrive=new COpendriveParamPoly3();
    else
    {
      error << "Unsupported or missing geometry in road " << this->name;
      throw CException(_HERE_,error.str());
    }
    geometry.opendrive->set_scale_factor(this->scale_factor);
    geometry.opendrive->load(*geom_info);
    if(geometry.opendrive->get_length()>this->min_road_length)
    {
      // create spline
      geometry.spline=new CG2Spline();
      start_pose=geometry.opendrive->get_start_pose();
      end_pose=geometry.opendrive->get_end_pose();
      geometry.opendrive->get_curvature(start_curvature,end_curvature);
      spline_start.x=start_pose.x;
      spline_start.y=start_pose.y;
      spline_start.heading=start_pose.heading;
      spline_start.curvature=start_curvature;
      geometry.spline->set_start_point(spline_start);
      spline_end.x=end_pose.x;
      spline_end.y=end_pose.y;
      spline_end.heading=end_pose.heading;
      spline_end.curvature=end_curvature;
      geometry.spline->set_end_point(spline_end);
      geometry.spline->generate(this->resolution,geometry.opendrive->get_length());
      this->geometries.push_back(geometry);
    }
  }
}

bool COpendriveRoadSegment::updated(segment_up_ref_t &refs)
{
  segment_up_ref_t::iterator updated_it;

  for(updated_it=refs.begin();updated_it!=refs.end();updated_it++)
    if(updated_it->second==this)
      return true;

  return false;
}

COpendriveRoadSegment *COpendriveRoadSegment::get_original_segment(segment_up_ref_t &segment_refs)
{
  segment_up_ref_t::iterator updated_it;

  for(updated_it=segment_refs.begin();updated_it!=segment_refs.end();updated_it++)
    if(updated_it->second==this)
      return updated_it->first;

  return NULL;
}

void COpendriveRoadSegment::update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs)
{
  std::vector<COpendriveRoadSegment *>::iterator seg_it;
  std::map<int,COpendriveLane *>::iterator lane_it;
  std::vector<COpendriveRoadNode *>::iterator node_it;

  if(this->updated(segment_refs))
  {
    for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();lane_it++)
    {
      if(lane_refs.find((*lane_it).second)!=lane_refs.end())
      {
        lane_it->second=lane_refs[lane_it->second];
        lane_it->second->update_references(segment_refs,lane_refs,node_refs);
      }
      else if(lane_it->second->updated(lane_refs))
        lane_it->second->update_references(segment_refs,lane_refs,node_refs);
    }
    for(node_it=this->nodes.begin();node_it!=this->nodes.end();node_it++)
      if(node_refs.find(*node_it)!=node_refs.end())
        (*node_it)=node_refs[*node_it];
    for(seg_it=this->connecting.begin();seg_it!=this->connecting.end();seg_it++)
      if(segment_refs.find(*seg_it)!=segment_refs.end())
        (*seg_it)=segment_refs[*seg_it];
    for(unsigned int i=0;i<this->signals.size();i++)
      this->signals[i]->update_references(segment_refs);
    for(unsigned int i=0;i<this->objects.size();i++)
      this->objects[i]->update_references(segment_refs);
  }
}

void COpendriveRoadSegment::clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs)
{
  std::vector<COpendriveRoadSegment *>::iterator seg_it;
  std::map<int,COpendriveLane *>::iterator lane_it;
  std::vector<COpendriveRoadNode *>::iterator node_it;

  if(this->updated(segment_refs))
  {
    for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();)
    {
      if(lane_it->second->updated(lane_refs))
      {
        lane_it->second->clean_references(segment_refs,lane_refs,node_refs);
        lane_it++;
      }
      else
        lane_it=this->lanes.erase(lane_it);
    }
    for(node_it=this->nodes.begin();node_it!=this->nodes.end();)
    {
      if(!(*node_it)->updated(node_refs))
        node_it=this->nodes.erase(node_it);
      else
        node_it++;
    }
    for(seg_it=this->connecting.begin();seg_it!=this->connecting.end();)
    {
      if(!(*seg_it)->updated(segment_refs))
        seg_it=this->connecting.erase(seg_it);
      else
        seg_it++;
    }
  }
}

void COpendriveRoadSegment::add_lanes(lanes::laneSection_type &lane_section)
{
  right::lane_iterator r_lane_it;
  left::lane_iterator l_lane_it;

  // right lanes
  if(lane_section.right().present())
  {
    for(r_lane_it=lane_section.right()->lane().begin();r_lane_it!=lane_section.right()->lane().end();r_lane_it++)
    {
      this->add_lane(*r_lane_it);
      this->num_right_lanes++;
    }
  }
  // left lanes
  if(lane_section.left().present())
  {
    for(l_lane_it=lane_section.left()->lane().begin();l_lane_it!=lane_section.left()->lane().end();l_lane_it++)
    {
      this->add_lane(*l_lane_it);
      this->num_left_lanes++;
    }
  }
}

void COpendriveRoadSegment::add_lane(const ::lane &lane_info)
{
  std::map<int,COpendriveLane *>::iterator lane_it; 
  COpendriveLane *new_lane;
  unsigned int lane_index;
  bool exist=false;

  try{
    new_lane=new COpendriveLane();
    new_lane->set_resolution(this->resolution);
    new_lane->set_scale_factor(this->scale_factor);
    new_lane->load(lane_info,this);
    for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();lane_it++)
      if(lane_it->first==new_lane->get_id())
      {
        exist=true;
        break;
      }
    if(!exist)
      this->lanes[new_lane->get_id()]=new_lane;
    lane_index=parent_road->add_lane(new_lane);
    new_lane->set_index(lane_index);
  }catch(CException &e){
    this->free();
    if(!this->parent_road->remove_lane(new_lane))
      delete new_lane;
    throw e;
  }
}

void COpendriveRoadSegment::remove_lane(COpendriveLane *lane)
{
  std::map<int,COpendriveLane *>::iterator lane_it;
  std::vector<COpendriveLane *>::iterator other_lane_it;

  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();)
  {
    if(lane_it->second==lane)
    {
      if(lane_it->first>0)
        this->num_left_lanes--;
      else
        this->num_right_lanes--;
      // remove associated nodes
      for(unsigned int i=0;i<lane->nodes.size();i++)
        this->remove_node(lane->nodes[i]);
      // remove the reference from neightbour lanes
      if(lane_it->second->left_lane!=NULL)
        lane_it->second->left_lane->remove_lane(lane);
      if(lane_it->second->right_lane!=NULL)
        lane_it->second->right_lane->remove_lane(lane);
      // remove the reference from next lanes
      for(unsigned int i=0;i<lane_it->second->next.size();i++)
        lane_it->second->next[i]->remove_lane(lane);
      // remove the reference from prev lanes
      for(unsigned int i=0;i<lane_it->second->prev.size();i++)
        lane_it->second->prev[i]->remove_lane(lane);
      lane_it=this->lanes.erase(lane_it);
      break;
    }
    else
      lane_it++;
  }
}

void COpendriveRoadSegment::add_nodes(void)
{
  double lane_offset;

  for(unsigned int j=0;j<this->geometries.size();j++)
  {
    lane_offset=0.0;
    for(int i=-1;i>=-this->num_right_lanes;i--)
    {
      this->lanes[i]->set_offset(lane_offset);
      this->add_node(this->geometries[j],this->lanes[i]);
      lane_offset-=this->lanes[i]->width;
    }
  }
  for(unsigned int j=0;j<this->geometries.size();j++)
  {
    lane_offset=0.0;
    for(int i=1;i<=this->num_left_lanes;i++)
    {
      this->lanes[i]->set_offset(lane_offset);
      this->add_node(this->geometries[this->geometries.size()-1-j],this->lanes[i]);
      lane_offset+=this->lanes[i]->width;
    }
  }
}

void COpendriveRoadSegment::add_node(TOpendriveRoadSegmentGeometry &geometry,COpendriveLane *lane)
{
  TOpendriveTrackPose track_pose;
  double start_curvature,end_curvature;
  COpendriveRoadNode *new_node;
  TOpendriveWorldPose node_pose;
  unsigned int node_index;
  bool exist=false;

  try{
    new_node=new COpendriveRoadNode();
    new_node->set_resolution(this->resolution);
    new_node->set_scale_factor(this->scale_factor);
    // import start position
    track_pose.t=lane->get_center_offset();
    track_pose.heading=0.0;
    if(lane->get_id()<0)
    {
      track_pose.s=0.0;
      geometry.opendrive->get_world_pose(track_pose,node_pose);
    }
    else
    {
      track_pose.s=geometry.opendrive->get_length();
      geometry.opendrive->get_world_pose(track_pose,node_pose);
      node_pose.heading=normalize_angle(node_pose.heading+3.14159);
    }
    new_node->set_pose(node_pose);
    if(this->parent_road->node_exists_at(node_pose))
    {
      delete new_node;
      new_node=this->parent_road->get_node_at(node_pose);
    }
    else
    {
      node_index=this->parent_road->add_node(new_node);
      new_node->set_index(node_index);
    }
    geometry.opendrive->get_curvature(start_curvature,end_curvature);
    if(start_curvature>0)
      start_curvature=1.0/((1.0/start_curvature)-lane->get_center_offset());
    else
      start_curvature=1.0/((1.0/start_curvature)+lane->get_center_offset());
    if(end_curvature>0)
      end_curvature=1.0/((1.0/end_curvature)-lane->get_center_offset());
    else
      end_curvature=1.0/((1.0/end_curvature)+lane->get_center_offset());
    if(lane->get_id()<0)
      new_node->add_parent(lane,this,start_curvature,end_curvature);
    else
      new_node->add_parent(lane,this,-end_curvature,-start_curvature);
    for(unsigned int i=0;i<this->nodes.size();i++)
      if(this->nodes[i]==new_node)
      {
        exist=true;
        break;
      }
    if(!exist)
      this->nodes.push_back(new_node);
    lane->add_node(new_node);
  }catch(CException &e){
    this->free();
    if(!this->parent_road->remove_node(new_node))
      delete new_node;
    throw e;
  }
}

void COpendriveRoadSegment::add_empty_node(TOpendriveWorldPose &pose,COpendriveLane *lane)
{
  COpendriveRoadNode *new_node;
  unsigned int node_index;
  bool exist=false;

  try{
    new_node=new COpendriveRoadNode();
    new_node->set_resolution(this->resolution);
    new_node->set_scale_factor(this->scale_factor);
    new_node->set_pose(pose);
    node_index=this->parent_road->add_node(new_node);
    new_node->set_index(node_index);
    for(unsigned int i=0;i<this->nodes.size();i++)
      if(this->nodes[i]==new_node)
      {
        exist=true;
        break;
      }
    if(!exist)
      this->nodes.push_back(new_node);
    lane->add_node(new_node);
  }catch(CException &e){
    this->free();
    if(!this->parent_road->remove_node(new_node))
      delete new_node;
    throw e;
  }
}

void COpendriveRoadSegment::remove_node(COpendriveRoadNode *node)
{
  std::vector<COpendriveRoadNode *>::iterator it;

  for(it=this->nodes.begin();it!=this->nodes.end();)
  {
    if((*it)==node)
    {
      it=this->nodes.erase(it);
      break;
    }
    else
      it++;
  }
}

bool COpendriveRoadSegment::has_node(COpendriveRoadNode *node) const
{
  for(unsigned int i=0;i<this->nodes.size();i++)
    if(this->nodes[i]==node)
      return true;

  return false;
}

int COpendriveRoadSegment::get_pose_side(TOpendriveWorldPose &pose) const
{
  const COpendriveRoadNode *node;
  std::map<int,COpendriveLane *>::const_iterator it;
  std::stringstream error;
  double distance;

  node=&this->get_closest_node(pose,distance,3.14159);
  for(it=this->lanes.begin();it!=this->lanes.end();it++)
    if(it->second->has_node((COpendriveRoadNode *)node))
      return it->first;

  error << "Segment " << this->name << " does not include node " << node->get_index();
  throw CException(_HERE_,error.str());
}

void COpendriveRoadSegment::link_neightbour_lanes(lanes::laneSection_type &lane_section)
{
  center::lane_type center_lane;

  if(lane_section.center().lane().present())
  {
    center_lane=lane_section.center().lane().get();
    if(center_lane.roadMark().size()>1)
      throw CException(_HERE_,"Only one road mark supported for lane");
    else if(center_lane.roadMark().size()==0)
      this->set_center_mark(OD_MARK_NONE);
    else if(center_lane.roadMark().size()==1)
    {
      if(center_lane.roadMark().begin()->type1().present())
      {
        if(center_lane.roadMark().begin()->type1().get()=="none")
          this->set_center_mark(OD_MARK_NONE);
        else if(center_lane.roadMark().begin()->type1().get()=="solid")
          this->set_center_mark(OD_MARK_SOLID);
        else if(center_lane.roadMark().begin()->type1().get()=="broken")
          this->set_center_mark(OD_MARK_BROKEN);
        else if(center_lane.roadMark().begin()->type1().get()=="solid solid")
          this->set_center_mark(OD_MARK_SOLID_SOLID);
        else if(center_lane.roadMark().begin()->type1().get()=="solid broken")
          this->set_center_mark(OD_MARK_SOLID_BROKEN);
        else if(center_lane.roadMark().begin()->type1().get()=="broken solid")
          this->set_center_mark(OD_MARK_BROKEN_SOLID);
        else if(center_lane.roadMark().begin()->type1().get()=="broken broken")
          this->set_center_mark(OD_MARK_BROKEN_BROKEN);
        else
          this->set_center_mark(OD_MARK_NONE);
      }
      else
        this->set_center_mark(OD_MARK_NONE);
    }
  }
  this->link_neightbour_lanes();
}

void COpendriveRoadSegment::link_neightbour_lanes(void)
{
  for(int i=-this->num_right_lanes;i<0;i++)
  {
    if(this->lanes.find(i+1)!=this->lanes.end())// if the lane exists 
      this->lanes[i]->link_neightbour_lane(this->lanes[i+1]);
    else
      this->lanes[i]->set_left_lane(NULL,this->center_mark);
  }
  for(int i=this->num_left_lanes;i>0;i--)
  {
    if(this->lanes.find(i-1)!=this->lanes.end())// if the lane exists 
      this->lanes[i]->link_neightbour_lane(this->lanes[i-1]);
    else
      this->lanes[i]->set_left_lane(NULL,this->center_mark);
  }
}

void COpendriveRoadSegment::link_segment(COpendriveRoadSegment *segment)
{
  if(this->connects_to(segment) || segment->connects_to(this))
    return;
  this->connecting.push_back(segment);
  segment->connecting.push_back(this);
  // link lanes
  for(int i=-this->num_right_lanes;i<0;i++)
  {
    for(int j=i-1;j<=i+1;j++)
    {
      if(segment->lanes.find(j)!=segment->lanes.end())
      {
        if(j==i-1)
          this->lanes[i]->link_lane(segment->lanes[j],this->lanes[i]->right_mark,false,true,false);
        else if(j==i)
          this->lanes[i]->link_lane(segment->lanes[j],OD_MARK_NONE,false,true,true);
        else
          this->lanes[i]->link_lane(segment->lanes[j],this->lanes[i]->left_mark,false,true,false);
      }
    }
  }
  for(int i=1;i<=segment->num_left_lanes;i++)
  {
    for(int j=i-1;j<=i+1;j++)
    {
      if(this->lanes.find(j)!=this->lanes.end())
      {
        if(j==i-1)
          segment->lanes[i]->link_lane(this->lanes[j],segment->lanes[i]->right_mark,false,true,false);
        else if(j==i)
          segment->lanes[i]->link_lane(this->lanes[j],OD_MARK_NONE,false,true,true);
        else
          segment->lanes[i]->link_lane(this->lanes[j],segment->lanes[i]->left_mark,false,true,false);
      }
    }
  }
}

void COpendriveRoadSegment::link_segment(COpendriveRoadSegment *segment,int from,bool from_start,int to,bool to_start)
{
  if(!this->connects_to(segment))
    this->connecting.push_back(segment);
  if(!segment->connects_to(this))
    segment->connecting.push_back(this);
  // link lanes
  if(segment->lanes.find(to)!=segment->lanes.end())
  {
    if(this->lanes.find(from-1)!=this->lanes.end())
      this->lanes[from-1]->link_lane(segment->lanes[to],this->lanes[from-1]->right_mark,from_start,to_start,false);
    if(this->lanes.find(from)!=this->lanes.end())
      this->lanes[from]->link_lane(segment->lanes[to],OD_MARK_NONE,from_start,to_start,true);
    if(this->lanes.find(from+1)!=this->lanes.end())
      this->lanes[from+1]->link_lane(segment->lanes[to],this->lanes[from+1]->left_mark,from_start,to_start,false);
  }
/*
  if(segment.lanes.find(to-1)!=segment.lanes.end())
    if(this->lanes.find(from)!=this->lanes.end())
      this->lanes[from]->link_lane(segment.lanes[to-1],this->lanes[from]->right_mark,from_start,to_start,false);
  if(segment.lanes.find(to+1)!=segment.lanes.end())
    if(this->lanes.find(from)!=this->lanes.end())
      this->lanes[from]->link_lane(segment.lanes[to+1],this->lanes[from]->left_mark,from_start,to_start,false);
*/
}

bool COpendriveRoadSegment::connects_to(COpendriveRoadSegment* segment)
{
  for(unsigned int i=0;i<this->connecting.size();i++)
    if(this->connecting[i]->get_id()==segment->get_id())
      return true;

  return false;
}

bool COpendriveRoadSegment::connects_segments(COpendriveRoadSegment *segment1,COpendriveRoadSegment *segment2)
{
  if(this->connects_to(segment1) && this->connects_to(segment2))
    return true;
  else
    return false;
}

COpendriveRoadSegment *COpendriveRoadSegment::get_sub_segment(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,int node_side,TOpendriveWorldPose *start,TOpendriveWorldPose *end) const
{
  TOpendriveWorldPose *start_pose,*end_pose;
  std::map<int,COpendriveLane *>::iterator lane_it;
  lane_up_ref_t::iterator lane_ref_it;
  COpendriveLane *new_lane;
  std::vector<COpendriveRoadSegment *>::iterator seg_it;
  segment_up_ref_t new_segment_ref;
  COpendriveRoadSegment *new_segment;
  std::vector<COpendriveRoadNode *>::iterator node_it;
  node_up_ref_t::iterator node_ref_it;
  std::vector<TOpendriveRoadSegmentGeometry>::iterator geom_it;
  TPoint new_point;
  double length;

  if(start==NULL && end==NULL)
    return this->clone(new_node_ref,new_lane_ref,new_link_ref);
  new_segment=new COpendriveRoadSegment(*this);
  new_segment_ref[(COpendriveRoadSegment *)this]=new_segment;
  if(node_side<0)
  {
    start_pose=start;
    end_pose=end;
  }
  else
  {
    start_pose=end;
    end_pose=start;
  }
  /* get the sublanes */
  for(lane_it=new_segment->lanes.begin();lane_it!=new_segment->lanes.end();lane_it++)
  {
    new_lane=lane_it->second->get_sub_lane(new_node_ref,new_lane_ref,new_link_ref,start_pose,end_pose);
    new_lane_ref[lane_it->second]=new_lane;
  }
  new_segment->update_references(new_segment_ref,new_lane_ref,new_node_ref);
  // update geometry
  if(start_pose!=NULL)
  {
    for(geom_it=new_segment->geometries.begin();geom_it!=new_segment->geometries.end();)
    {
      length=geom_it->spline->find_closest_point(start_pose->x,start_pose->y,new_point);
      if(length<std::numeric_limits<double>::max())
      {
        geom_it->spline->set_start_point(new_point);
        geom_it->spline->generate(this->resolution);
        geom_it->opendrive->set_start_pose(*start_pose);
        geom_it->opendrive->set_max_s(geom_it->opendrive->get_max_s()-length);
        break;
      }
      else
        geom_it=new_segment->geometries.erase(geom_it);
    }
    for(unsigned int i=1;i<new_segment->geometries.size();i++)
    {
      new_segment->geometries[i].opendrive->set_min_s(new_segment->geometries[i].opendrive->get_min_s()-length);
      new_segment->geometries[i].opendrive->set_max_s(new_segment->geometries[i].opendrive->get_max_s()-length);
    }  
  }
  if(end_pose!=NULL)
  {
    for(geom_it=new_segment->geometries.begin();geom_it!=new_segment->geometries.end();)
    {
      length=geom_it->spline->find_closest_point(end_pose->x,end_pose->y,new_point);
      if(length<std::numeric_limits<double>::max())
      {
        geom_it->spline->set_end_point(new_point);
        geom_it->spline->generate(this->resolution);
        geom_it->opendrive->set_max_s(geom_it->opendrive->get_min_s()+length);;
        geom_it=new_segment->geometries.erase(++geom_it,new_segment->geometries.end());
        break;
      }
      else
        geom_it++;
    }
  }

  return new_segment; 
}

COpendriveRoadSegment *COpendriveRoadSegment::clone(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref) const
{
  std::map<int,COpendriveLane *>::iterator lane_it;
  lane_up_ref_t::iterator lane_ref_it;
  COpendriveLane *new_lane;
  segment_up_ref_t new_segment_ref;
  COpendriveRoadSegment *new_segment;
  std::vector<COpendriveRoadNode *>::iterator node_it;
  node_up_ref_t::iterator node_ref_it;

  new_segment=new COpendriveRoadSegment(*this);
  new_segment_ref[(COpendriveRoadSegment *)this]=new_segment;
  /* get the sublanes */
  for(lane_it=new_segment->lanes.begin();lane_it!=new_segment->lanes.end();lane_it++)
  {
    new_lane=lane_it->second->clone(new_node_ref,new_lane_ref,new_link_ref);
    new_lane_ref[lane_it->second]=new_lane;
  }
  new_segment->update_references(new_segment_ref,new_lane_ref,new_node_ref);

  return new_segment;
}

const COpendriveRoadSegment &COpendriveRoadSegment::get_next_segment(int &side,bool &valid) const
{
  std::vector<COpendriveRoadSegment *> segment_candidates;
  std::vector<int> side_candidates;
  bool already_present;

  if(side<0)
  {
    for(unsigned int i=1;i<=this->get_num_right_lanes();i++)
    {
      const COpendriveLane &lane=this->get_lane(-i);
      if(lane.get_num_next_lanes()>1)
        throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
      if(lane.get_num_next_lanes()==1)
      {
        already_present=false;
        for(unsigned int k=0;k<segment_candidates.size();k++)
          if(segment_candidates[k]==lane.get_next_lane(0).segment)
          {
            already_present=true;
            break;
          }
        if(!already_present)
        {
          segment_candidates.push_back(lane.get_next_lane(0).segment);
          side_candidates.push_back(lane.get_next_lane(0).get_id());
        }
      }
    }
  }
  else
  {
    for(unsigned int i=1;i<=this->get_num_left_lanes();i++)
    {
      const COpendriveLane &lane=this->get_lane(i);
      if(lane.get_num_next_lanes()>1)
        throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
      if(lane.get_num_next_lanes()==1)
      {
        already_present=false;
        for(unsigned int k=0;k<segment_candidates.size();k++)
          if(segment_candidates[k]==lane.get_next_lane(0).segment)
          {
            already_present=true;
            break;
          }
        if(!already_present)
        {
          segment_candidates.push_back(lane.get_next_lane(0).segment);
          side_candidates.push_back(lane.get_next_lane(0).get_id());
        }
      }
    }
  }
  if(segment_candidates.size()==0)
  {
    valid=false;
    return *this;
  }
  else if(segment_candidates.size()>1)
    throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
  else
  {
    valid=true;
    side=side_candidates[0];
    return *segment_candidates[0];
  }
}

const COpendriveRoadSegment &COpendriveRoadSegment::get_prev_segment(int &side,bool &valid) const
{
  std::vector<COpendriveRoadSegment *> segment_candidates;
  std::vector<int> side_candidates;
  bool already_present;

  if(side<0)
  {
    for(unsigned int i=1;i<=this->get_num_right_lanes();i++)
    {
      const COpendriveLane &lane=this->get_lane(-i);
      if(lane.get_num_prev_lanes()>1)
        throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
      if(lane.get_num_prev_lanes()==1)
      {
        already_present=false;
        for(unsigned int k=0;k<segment_candidates.size();k++)
          if(segment_candidates[k]==lane.get_prev_lane(0).segment)
          {
            already_present=true;
            break;
          }
        if(!already_present)
        {
          segment_candidates.push_back(lane.get_prev_lane(0).segment);
          side_candidates.push_back(lane.get_prev_lane(0).get_id());
        }
      }
    }
  }
  else
  {
    for(unsigned int i=1;i<=this->get_num_left_lanes();i++)
    {
      const COpendriveLane &lane=this->get_lane(i);
      if(lane.get_num_prev_lanes()>1)
        throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
      if(lane.get_num_prev_lanes()==1)
      {
        already_present=false;
        for(unsigned int k=0;k<segment_candidates.size();k++)
          if(segment_candidates[k]==lane.get_prev_lane(0).segment)
          {
            already_present=true;
            break;
          }
        if(!already_present)
        {
          segment_candidates.push_back(lane.get_prev_lane(0).segment);
          side_candidates.push_back(lane.get_prev_lane(0).get_id());
        }
      }
    }
  }
  if(segment_candidates.size()==0)
  {
    valid=false;
    return *this;
  }
  else if(segment_candidates.size()>1)
    throw CException(_HERE_,"Road has multiple paths and no unique subroad exists");
  else
  {
    valid=true;
    side=side_candidates[0];
    return *segment_candidates[0];
  }
}


void COpendriveRoadSegment::load(OpenDRIVE::road_type &road_info)
{
  std::map<int,COpendriveLane *>::const_iterator lane_it;
  signals::signal_iterator signal_it;
  objects::object_iterator object_it;
  lanes::laneSection_iterator lane_section;
  COpendriveSignal *new_signal;
  COpendriveObject *new_object;
  std::stringstream error;

  this->free();
  this->set_name(road_info.name().get());
  this->set_id(std::stoi(road_info.id().get()));
  // only one lane section is supported
  if(road_info.lanes().laneSection().size()<1)
  {
    error << "No lane sections defined for segment " << this->name;
    throw CException(_HERE_,error.str());
  }
  else if(road_info.lanes().laneSection().size()>1)
  {
    error << "Segment " << this->name << " has more than one lane section";
    throw CException(_HERE_,error.str());
  }
  else
    lane_section=road_info.lanes().laneSection().begin();

  try{
    // add lanes
    this->add_lanes(*lane_section);
    // add geometries
    this->add_geometries(road_info);
    // add road nodes
    this->add_nodes();
    // link lanes
    this->link_neightbour_lanes(*lane_section);
    // add road signals
    if(road_info.signals().present())
    {
      for(signal_it=road_info.signals().get().signal().begin();signal_it!=road_info.signals().get().signal().end();++signal_it)
      {
        new_signal=new COpendriveSignal();
        new_signal->load(*signal_it,this);
        new_signal->set_scale_factor(this->scale_factor);
        this->signals.push_back(new_signal);
      }
    }
    // add road objects
    if(road_info.objects().present())
    {
      for(object_it=road_info.objects().get().object().begin();object_it!=road_info.objects().get().object().end();++object_it)
      {
        new_object=new COpendriveObject();
        new_object->load(*object_it,this);
        new_object->set_scale_factor(this->scale_factor);
        this->objects.push_back(new_object);
      }
    }
  }catch(CException &e){
    this->free();
    throw e;
  }
}

std::string COpendriveRoadSegment::get_name(void) const
{
  return this->name;
}

unsigned int COpendriveRoadSegment::get_id(void) const
{
  return this->id;
}

unsigned int COpendriveRoadSegment::get_num_right_lanes(void) const
{
  return this->num_right_lanes;
}

unsigned int COpendriveRoadSegment::get_num_left_lanes(void) const
{
  return this->num_left_lanes;
}

const COpendriveLane &COpendriveRoadSegment::get_lane(int index) const
{
  std::map<int,COpendriveLane *>::const_iterator lane_it;
  std::stringstream error;

  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();++lane_it)
  {
    if(lane_it->second->get_id()==index)
      return *lane_it->second;
  }

  error << "Invalid lane index (" << index << ") for segment " << this->name;
  throw CException(_HERE_,error.str());
}

unsigned int COpendriveRoadSegment::get_num_nodes(void) const
{
  return this->nodes.size();
}

const COpendriveRoadNode &COpendriveRoadSegment::get_node(unsigned int index) const
{
  std::stringstream error;

  if(index>=0 && index<this->nodes.size())
    return *this->nodes[index];
  else
  {
    error << "Invalid node index (" << index << ") for segment " << this->name;
    throw CException(_HERE_,error.str());
  }
}

const COpendriveRoad &COpendriveRoadSegment::get_parent_road(void) const
{
  return *this->parent_road;
}

unsigned int COpendriveRoadSegment::get_num_signals(void) const
{
  return this->signals.size();  
}

const COpendriveSignal &COpendriveRoadSegment::get_signal(unsigned int index) const
{
  std::stringstream error;

  if(index>=0 && index<this->signals.size())
    return *this->signals[index];
  else
  {
    error << "Invalid signal index (" << index << ") for segment " << this->name;
    throw CException(_HERE_,error.str());
  }
}

unsigned int COpendriveRoadSegment::get_num_objects(void) const
{
  return this->objects.size();
}

const COpendriveObject &COpendriveRoadSegment::get_object(unsigned int index) const
{
  std::stringstream error;

  if(index>=0 && index<this->objects.size())
    return *this->objects[index];
  else
  {
    error << "Invalid object index (" << index << ") for segment " << this->name;
    throw CException(_HERE_,error.str());
  }
}

unsigned int COpendriveRoadSegment::get_num_connecting(void) const
{
  return this->connecting.size();
}

const COpendriveRoadSegment &COpendriveRoadSegment::get_connecting(unsigned int index) const
{
  std::stringstream error;

  if(index>=0 && index<this->connecting.size())
    return *this->connecting[index];
  else
  {
    error << "Invalid connecting segment index (" << index << ") for segment " << this->name;
    throw CException(_HERE_,error.str());
  }
}

double COpendriveRoadSegment::get_length(void) const
{
  double length=0.0;

  for(unsigned int i=0;i<this->geometries.size();i++)
    length+=this->geometries[i].spline->get_length();

  return length;
}

TOpendriveWorldPose COpendriveRoadSegment::get_pose_at(double length) const
{
  TOpendriveWorldPose world_pose{0};
  TPoint spline_pose;

  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    if((length-this->geometries[i].spline->get_length())<0.0)
    {
      spline_pose=this->geometries[i].spline->evaluate(length);
      world_pose.x=spline_pose.x;
      world_pose.y=spline_pose.y;
      world_pose.heading=spline_pose.heading;
      return world_pose;
    }
    else
      length-=this->geometries[i].spline->get_length();
  }

  return world_pose;
}

double COpendriveRoadSegment::get_curvature_at(double length) const
{
  double curvature;
  TPoint spline_pose;

  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    if((length-this->geometries[i].spline->get_length())<0.0)
    {
      spline_pose=this->geometries[i].spline->evaluate(length);
      curvature=spline_pose.curvature;
      return curvature;
    }
    else
      length-=this->geometries[i].spline->get_length();
  }

  return 0.0;
}

TOpendriveLocalPose COpendriveRoadSegment::transform_to_local_pose(const TOpendriveTrackPose &track)
{
  TOpendriveTrackPose pose;
  TOpendriveLocalPose local;
  std::stringstream error;

  if(track.s<0.0)
  {
    error << "Invalid track pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ")";
    throw CException(_HERE_,error.str());
  }
  pose=track;
  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    if((pose.s-this->geometries[i].opendrive->get_length())<=this->resolution)
    {
      if(!this->geometries[i].opendrive->get_local_pose(pose,local))
      {
        error << "Impossible to transform to local coordinates pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ") in segment " << this->name;
        throw CException(_HERE_,error.str());
      }
      return local;
    }
    else
      pose.s-=this->geometries[i].opendrive->get_length();
  }

  error << "Track pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ") does not belong to segment " << this->name;
  throw CException(_HERE_,error.str());
}

TOpendriveWorldPose COpendriveRoadSegment::transform_to_world_pose(const TOpendriveTrackPose &track)
{
  TOpendriveTrackPose pose;
  TOpendriveWorldPose world;
  std::stringstream error;

  if(track.s<0.0)
  {
    error << "Invalid track pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ")";
    throw CException(_HERE_,error.str());
  }
  pose=track;
  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    if((pose.s-this->geometries[i].opendrive->get_length())<=this->resolution)
    {
      if(!this->geometries[i].opendrive->get_world_pose(pose,world))
      {
        error << "Impossible to transform to world coordinates pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ") in segment " << this->name;
        throw CException(_HERE_,error.str());
      }
      return world;
    }
    else
      pose.s-=this->geometries[i].opendrive->get_length();
  }

  error << "Track pose (s=" << track.s << ",t=" << track.t << ",heading=" << track.heading << ") does not belong to segment " << this->name;
  throw CException(_HERE_,error.str());
}

unsigned int COpendriveRoadSegment::get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol) const
{
  double dist,min_dist=std::numeric_limits<double>::max(),length;
  TOpendriveWorldPose closest_pose;
  unsigned int closest_index=-1;

  for(unsigned int i=0;i<this->nodes.size();i++)
  {
    length=this->nodes[i]->get_closest_pose(pose,closest_pose,false,angle_tol);
    if(length<std::numeric_limits<double>::max())
    {
      dist=sqrt(pow(closest_pose.x-pose.x,2.0)+pow(closest_pose.y-pose.y,2.0));
      if(dist<min_dist)
      {
        min_dist=dist;
        closest_index=i;
        distance=length;
      }
    }
  }

  return closest_index;
}

const COpendriveRoadNode &COpendriveRoadSegment::get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol) const
{
  unsigned int closest_index;

  closest_index=this->get_closest_node_index(pose,distance,angle_tol);
  if(closest_index==(unsigned int)-1)
    throw CException(_HERE_,"Impossible to find the closest node");
  return *this->nodes[closest_index];
}

int COpendriveRoadSegment::get_closest_lane_id(TOpendriveWorldPose &pose,double &distance,double angle_tol) const
{
  double dist,min_dist=std::numeric_limits<double>::max(),length;
  TOpendriveWorldPose closest_pose;
  int closest_id=0;
  std::map<int,COpendriveLane *>::const_iterator lane_it;

  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();lane_it++)
  {
    length=lane_it->second->get_closest_pose(pose,closest_pose,angle_tol);
    if(length<std::numeric_limits<double>::max())
    {
      dist=sqrt(pow(closest_pose.x-pose.x,2.0)+pow(closest_pose.y-pose.y,2.0));
      if(dist<min_dist)
      {
        min_dist=dist;
        closest_id=lane_it->first;
        distance=length;
      }
    }
  }

  return closest_id;
}

const COpendriveLane &COpendriveRoadSegment::get_closest_lane(TOpendriveWorldPose &pose,double &distance,double angle_tol) const
{
  std::map<int,COpendriveLane *>::const_iterator lane_it;

  int closest_id=this->get_closest_lane_id(pose,distance,angle_tol);
  if(closest_id==0)
    throw CException(_HERE_,"Impossible to find the closest lane");
  for(lane_it=this->lanes.begin();lane_it!=this->lanes.end();lane_it++)
    if(lane_it->first==closest_id)
      return *lane_it->second;

  throw CException(_HERE_,"Impossible to find the closest lane");
}

double COpendriveRoadSegment::get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol) const
{
  double dist,min_dist=std::numeric_limits<double>::max(),distance,length;
  unsigned int closest_index=-1;
  TPoint closest_spline_point;

  closest_pose.x=std::numeric_limits<double>::max();
  closest_pose.y=std::numeric_limits<double>::max();
  closest_pose.heading=std::numeric_limits<double>::max();
  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    length=this->geometries[i].spline->find_closest_point(pose.x,pose.y,closest_spline_point);
    if(length<std::numeric_limits<double>::max())
    {
      dist=sqrt(pow(closest_spline_point.x-pose.x,2.0)+pow(closest_spline_point.y-pose.y,2.0));
      if(dist<min_dist)
      {
        min_dist=dist;
        closest_index=i;
        closest_pose.x=closest_spline_point.x;
        closest_pose.y=closest_spline_point.y;
        closest_pose.heading=normalize_angle(closest_spline_point.heading);
        distance=length;
      }
    }
  }

  if(closest_index!=(unsigned int)-1)
  {
    for(unsigned int i=0;i<closest_index;i++)
      distance+=this->geometries[i].spline->get_length();
  }
  else
    distance=std::numeric_limits<double>::max();

  return distance;
}

std::ostream& operator<<(std::ostream& out, COpendriveRoadSegment &segment)
{
  out << "Road " << segment.get_name() << " (" << segment.get_id() << ")" << std::endl;
  out << "  Connecting road segments: " << segment.connecting.size() << std::endl;
  for(unsigned int i=0;i<segment.connecting.size();i++)
    out << "    " << segment.connecting[i]->get_name() << std::endl;
  out << "  Number of nodes: " << segment.get_num_nodes() << std::endl;
  for(unsigned int i=0;i<segment.get_num_nodes();i++)
    out << "    " << segment.get_node(i).get_index() << std::endl;
  out << "  Number of right lanes: " << segment.num_right_lanes << std::endl;
  for(int i=-1;i>=-segment.num_right_lanes;i--)
    out << *segment.lanes[i];
  out << "  Number of left lanes: " << segment.num_left_lanes << std::endl;
  for(int i=1;i<=segment.num_left_lanes;i++)
    out << *segment.lanes[i];
  out << "  Number of signals: " << segment.signals.size() << std::endl;
  for(unsigned int i=0;i<segment.signals.size();i++)
    out << *segment.signals[i];
  out << "  Number of objects: " << segment.objects.size() << std::endl;
  for(unsigned int i=0;i<segment.objects.size();i++)
    out << *segment.objects[i];

  return out;
}

COpendriveRoadSegment::~COpendriveRoadSegment()
{
  this->free();
  this->name="";
  this->id=-1;
  this->center_mark=OD_MARK_NONE;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->min_road_length=DEFAULT_MIN_ROAD_LENGTH;
}

